import typer
import readline  # noqa

from beam import BEAM_CONFIG
from beam.decorators import with_config
from beam import console
from beam.clients.api import BeamApiClient
from beam.clients.runner import RunnerClient


@with_config
def up():
    """
    Push local changes to remote sandbox store.
    """

    console.log(f"Loading sandbox with compute settings: {BEAM_CONFIG.compute}")



from beam.config import BeamConfigV1
from beam import log
from rich.console import Console
from marshmallow.exceptions import ValidationError

BEAM_CONFIG: BeamConfigV1 = BeamConfigV1()
console = Console()


def load_config():
    try:
        BEAM_CONFIG.load()
    except FileNotFoundError:
        log.die(
            "No configuration file found - make sure this directory has a 'config.yaml' file present."
        )
    except ValidationError as exc:
        log.die(f"Invalid config file: {exc}")



import requests

from typing import Optional
from beam.constants import (
    REQUESTS_TIMEOUT,
    BACKEND_BASE_URL,
)

TOKEN = "c3a9d1422cc0d4847e02f66ade0e9d4e1174378f45b91e09bde0e49063efddc3b3bb02dee19d4f34a03f0f014a3cfee1ca16409beb3eb1a568675435a4b90b0e"  # TODO: replace with API key based auth


class BeamApiClient:
    @classmethod
    def _post(cls, *, route: str, body):
        body = {k: v for k, v in body.items() if v is not None}
        res = requests.post(
            f"{BACKEND_BASE_URL}/{route}",
            headers={"Authorization": f"identity {TOKEN}"},
            json=body,
            timeout=REQUESTS_TIMEOUT,
        )
        res.raise_for_status()
        return res.json()

    @classmethod
    def down(cls, *, sandbox_id: Optional[str] = None):
        return cls._post(route="sandbox/beam/down", body={"sandbox_id": sandbox_id})

    @classmethod
    def up(cls, *, sandbox_id: Optional[str] = None):
        return cls._post(route="sandbox/beam/up", body={"sandbox_id": sandbox_id})

    @classmethod
    def start(cls, *, sandbox_name: str):
        return cls._post(
            route="sandbox/beam/start", body={"sandbox_name": sandbox_name}
        )

    @classmethod
    def init(cls, *, sandbox_id: Optional[str] = None):
        return cls._post(route="sandbox/beam/init", body={"sandbox_id": sandbox_id})



import sys
import os
import datetime

from rich import print


OUTPUT_PREFIX = ""


def _timestamp(str):
    t = datetime.datetime.now()
    return f"{t.hour:02d}:{t.minute:02d}:{t.second:02d}  {str}"


def warn(str):
    print(f"{OUTPUT_PREFIX}{_timestamp(str)}")  # noqa


def debug(str):
    if int(os.getenv("BEAM_DEBUG", 0)) != 0:
        print(f"{OUTPUT_PREFIX}{_timestamp(str)}")  # noqa


def die(str):
    print(f"{OUTPUT_PREFIX}{_timestamp(str)}")  # noqa
    sys.exit(1)


def info(str):
    print(f"{OUTPUT_PREFIX}{_timestamp(str)}")  # noqa



class CommandTypes:
    READY = "READY"
    RUN_COMMAND = "RUN_COMMAND"
    STDIN = "STDIN"
    TYPES = (
        (READY, "ready"),
        (RUN_COMMAND, "run_command"),
        (STDIN, "stdin"),
    )



import yaml

from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Any
from dataclasses_json import DataClassJsonMixin
from yamldataclassconfig.config import YamlDataClassConfig
from yamldataclassconfig import create_file_path_field
from beam import constants


class Runtimes:
    PYTHON_38 = "python3.8"


@dataclass
class ComputeConfig(DataClassJsonMixin):
    gpu: int
    cpu: int
    mem: str


@dataclass
class BeamConfigV1(YamlDataClassConfig):
    FILE_PATH: Path = create_file_path_field(
        Path(__file__).parent.cwd() / "config.yaml"
    )
    name: Optional[str] = None
    compute: ComputeConfig = field(default=ComputeConfig(gpu=0, cpu=4, mem="8Gi"))
    apt: List[str] = field(default_factory=lambda: [])
    runtime: str = Runtimes.PYTHON_38
    packages: List[str] = field(default_factory=lambda: [])


def write_configuration(*, sandbox: Any, path: str) -> bool:
    compute_settings = sandbox.get("compute_settings", {})
    apt = sandbox.get("apt", [])  # TODO: add in system level dependencies
    packages = sandbox.get("requirements", {})
    runtime = sandbox.get("python_runtime", Runtimes.PYTHON_38).lower()

    cpu = compute_settings.get("cpus", constants.DEFAULT_CONFIG_CPU)
    memory = compute_settings.get("gpus", constants.DEFAULT_CONFIG_MEMORY)
    gpu = compute_settings.get("gpus", constants.DEFAULT_CONFIG_GPU)

    config = {
        "name": sandbox["name"],
        "compute": {
            "cpu": cpu,
            "mem": memory,
            "gpu": gpu,
        },
        "apt": apt,
        "runtime": runtime,
        "packages": [f"{req}=={version}" for req, version in packages.items()],
    }

    with open(f"{path}/config.yaml", "w") as f:
        yaml.dump(config, f, default_flow_style=False, sort_keys=False)

    return True



import typer

from beam.commands import up
from beam.commands import down
from beam.commands import start


app = typer.Typer()
app.command()(up.up)
app.command()(down.down)
app.command()(start.start)


if __name__ == "__main__":
    app()



import subprocess
import os
import signal
import shlex

from typing import List


class ProcessManager:
    def __init__(self) -> None:
        self.proc = None

    def start(self, process_list: List[str]) -> None:
        # Kill existing process if anything is running
        if self.proc is not None:
            self.stop()

        self.proc = subprocess.Popen(
            process_list,
            preexec_fn=os.setsid,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

    def start_from_string(self, process_string: str):
        return self.start(shlex.split(process_string))

    def stop(self) -> None:
        if self.proc is None:
            return

        try:
            os.kill(self.proc.pid, signal.SIGINT)
        except OSError:
            pass

        self.proc = None

    def is_running(self) -> bool:
        if self.proc is None:
            return False
        else:
            if self.proc.poll() is None:
                return True
            else:
                return False

    def __del__(self) -> None:
        self.stop()  # Stop if class is destroyed



import json
import asyncio
import shlex

from websockets import connect
from websockets.exceptions import InvalidStatusCode
from typing import Any, List
from rich.progress import Progress
from beam.commands import CommandTypes
from beam import BEAM_CONFIG, console


class Websocket:
    def __init__(self, uri: str) -> None:
        self.uri = uri

    async def __aenter__(self):
        self._conn = connect(self.uri)
        self.websocket = await self._conn.__aenter__()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self._conn.__aexit__(*args, **kwargs)

    async def send(self, message):
        await self.websocket.send(message)

    async def receive(self):
        return await self.websocket.recv()


async def response_handler(
    *, progress: Any, command: str, ws: Websocket, rx_timeout: float = 0.1
):
    response = {}

    if command == CommandTypes.READY:
        try:
            response = await asyncio.wait_for(ws.receive(), timeout=rx_timeout)
            if response:
                response = json.loads(response)

        except asyncio.exceptions.TimeoutError:
            pass

    elif command == CommandTypes.RUN_COMMAND:
        exit_code = None

        while exit_code is None:
            try:
                response = await asyncio.wait_for(ws.receive(), timeout=rx_timeout)
                if response:
                    response = json.loads(response)

                stdout = response.get("stdout", "")
                exit_code = response.get("exit_code")

                if stdout is not None and stdout != "":
                    progress.print(stdout)

            except asyncio.exceptions.TimeoutError:
                pass

    return response


class RunnerClient:
    def __init__(self, *, sandbox_session_id: str) -> None:
        self.sandbox_session_id = sandbox_session_id

        # self.ws_uri = "ws://localhost:8008"
        namespace = "luke-lombardi"
        self.ws_uri = (
            f"wss://sandbox-{namespace}.slai.okteto.dev/beam/{self.sandbox_session_id}/"
        )

    async def _send_command(
        self,
        *,
        payload: Any,
    ):
        response = {}

        try:
            async with Websocket(self.ws_uri) as ws:
                await ws.send(json.dumps(payload))
                response = await response_handler(
                    progress=self.progress, command=payload.get("command"), ws=ws
                )
        except InvalidStatusCode:
            pass

        return response

    def _wait_for_ready(self) -> bool:
        ready = False
        payload = {
            "command": CommandTypes.READY,
            "token": "fake_token",
        }

        while not ready:
            response = asyncio.run(self._send_command(payload=payload))
            ready = response.get("ready") is True

            if ready:
                break

        return ready

    def _create_environment(self):
        # TODO: replace with requirements from config
        requirements = {"torch": "1.10.2"}
        # BEAM_CONFIG.packages[0]

        payload = {
            "command": CommandTypes.RUN_COMMAND,
            "token": "fake_token",
            "cmd": [
                "python",
                "-m",
                "runner.develop.create",
                f"{json.dumps(requirements)}",
            ],
            "env": {},
        }

        response = asyncio.run(self._send_command(payload=payload))
        exit_code = response.get("exit_code")

        return exit_code

    def _run_cmd(self, cmd: List[str]):
        payload = {
            "command": CommandTypes.RUN_COMMAND,
            "token": "fake_token",
            "cmd": cmd,
            "env": {},
        }

        response = asyncio.run(self._send_command(payload=payload))
        exit_code = response.get("exit_code")

        return exit_code

    def initialize(self) -> bool:
        with Progress(transient=True) as progress:
            self.progress = progress
            progress.add_task("Waiting for runner to start...", total=None)
            ready = self._wait_for_ready()

        if not ready:
            return False

        with Progress(transient=True) as progress:
            self.progress = progress
            progress.add_task("Creating environment...", total=None)
            exit_code = self._create_environment()

        return exit_code == 0

    def run(self, cmd: str) -> bool:
        cmd = shlex.split(cmd)
        exit_code = self._run_cmd(cmd=cmd)
        return exit_code == 0



import readline  # noqa

from operator import itemgetter
from pathlib import Path
from beam.util import download_files
from beam.config import write_configuration
from beam import console
from beam.clients.api import BeamApiClient


def down(sandbox_id: str) -> None:
    """
    Download a sandbox for local development.
    """

    console.log(f"Pulling sandbox: '{sandbox_id}'")

    try:
        sandbox, download_urls = itemgetter("sandbox", "download_urls")(
            BeamApiClient.down(sandbox_id=sandbox_id)
        )
    except KeyError:
        console.log(f"[red]Unable to retrieve sandbox: '{sandbox_id}'")
        return

    name = sandbox["name"]

    try:
        Path(f"{name}").mkdir()
    except FileExistsError:
        console.log(f"[red]A folder already exists with name: '{name}'")
        return

    # Download sandbox contents
    if not download_files(download_urls, path=name):
        console.log(f"[red]Failed to download files.")
        return

    # Write configuration file
    if not write_configuration(sandbox=sandbox, path=name):
        console.log(f"[red]Failed to write configuration.")
        return

    console.log(f"Downloaded sandbox, run 'beam up' in './{name}'")



import os

REQUESTS_TIMEOUT = 120
BACKEND_BASE_URL = "http://localhost:8080"  # TODO: replace with dynamic API url
DEFAULT_CONFIG_DIRECTORY = os.path.expanduser("~/.beam")
DEFAULT_SYNCTHING_LOCAL_PORT = 32768
DEFAULT_SYNCTHING_REMOTE_PORT = 443
DEFAULT_CONFIG_CPU = 8
DEFAULT_CONFIG_MEMORY = "32Gi"
DEFAULT_CONFIG_GPU = 0



from functools import wraps
from pathlib import Path
from os import path
from beam import load_config


def with_config(callback):
    @wraps(callback)
    def wrapper(*args, **kwargs):
        load_config()
        return callback(*args, **kwargs)

    return wrapper


def requires_credentials(callback, *outer_args, **outer_kwargs):
    def wrapper(*args, **kwargs):
        profile_name = outer_kwargs.get("profile", "default")
        return callback(*args, **kwargs)

    return wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['beam', 'beam.clients', 'beam.clients.syncthing', 'beam.commands']

package_data = \
{'': ['*']}

install_requires = \
['python-syncthing-client>=0.1.1,<0.2.0',
 'requests>=2.28.1,<3.0.0',
 'rich>=12.5.1,<13.0.0',
 'typer>=0.6.1,<0.7.0',
 'websockets>=10.3,<11.0',
 'yamldataclassconfig>=1.5.0,<2.0.0']

entry_points = \
{'console_scripts': ['beam = beam.cli:app']}

setup_kwargs = {
    'name': 'beam-cli',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Luke Lombardi',
    'author_email': 'luke@slai.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0.0',
}


setup(**setup_kwargs)



import requests
import shutil
import os


from http import HTTPStatus
from typing import Callable, Optional, Tuple
from itertools import repeat
from urllib.error import HTTPError
from multiprocessing.pool import ThreadPool
from rich.progress import Progress

DOWNLOAD_THREAD_COUNT = 8


def download_files(download_urls: dict, path: Optional[str] = None) -> bool:
    result_map = _async_pool(download_urls, _get_object, path=path)

    for _, success in result_map.items():
        if not success:
            return False

    return True


def safe_open(path: str, mode: str = "w"):
    """Open "path" for writing, creating any parent directories as needed."""
    os.makedirs(os.path.dirname(path), exist_ok=True)
    return open(path, mode)


def _async_pool(objects: dict, callback: Callable, path: Optional[str] = None) -> dict:
    result_map = {}

    with Progress() as progress:
        task = progress.add_task(
            "[green]Downloading sandbox files...", total=len(objects.keys())
        )

        for result in ThreadPool(DOWNLOAD_THREAD_COUNT).starmap(
            callback, zip(objects.items(), repeat(path))
        ):
            progress.advance(task)
            result_map[result[0]] = result[1]

    return result_map


def _get_object(object: Tuple, path: Optional[str] = None) -> Tuple:
    filename = object[0]
    url = object[1]

    if path is None:
        path = os.getcwd()

    try:
        response = requests.get(url, stream=True)
        response.raise_for_status()

        if response.status_code == HTTPStatus.OK:
            with safe_open(f"{path}/{filename}", "wb") as f:
                response.raw.decode_content = True
                shutil.copyfileobj(response.raw, f)

    except (HTTPError, BaseException) as exc:
        return (filename, False)

    return (filename, True)



import typer
import readline  # noqa

from operator import itemgetter
from beam import BEAM_CONFIG
from beam.decorators import with_config
from beam import console
from beam.clients.api import BeamApiClient
from beam.clients.runner import RunnerClient
from beam.clients.syncthing import SyncthingClient


@with_config
def start():
    """
    Connect to a remote sandbox environment and start working.
    """
    name = BEAM_CONFIG.name
    console.log(
        f"Starting sandbox '{name}' with compute settings: {BEAM_CONFIG.compute}"
    )

    try:
        sandbox_session = itemgetter("sandbox_session")(
            BeamApiClient.start(sandbox_name=name)
        )
    except KeyError:
        console.log(f"[red]Unable to create sandbox session: '{name}'")
        return

    runner_client = RunnerClient(sandbox_session_id=sandbox_session["id"])
    if not runner_client.initialize():
        console.log(f"[red]Unable to connect to sandbox: '{name}'")

    try:
        syncthing_client = SyncthingClient(sandbox_session_id=sandbox_session["id"])
    except RuntimeError:
        console.log("[red]Unable to connect to syncthing server.")
        return

    while True:
        cmd = typer.prompt("#")
        runner_client.run(cmd)



import time
import os

from typing import Union, Tuple
from syncthing import Syncthing, SyncthingError
from beam.processmanager import ProcessManager
from beam.constants import (
    DEFAULT_CONFIG_DIRECTORY,
    DEFAULT_SYNCTHING_LOCAL_PORT,
    DEFAULT_SYNCTHING_REMOTE_PORT,
)
from beam.clients.syncthing.templates import (
    generate_folder_template,
    generate_folder_device_template,
    generate_remote_device_template,
    generate_local_device_template,
    DEFAULT_REMOTE_DEVICE_NAME,
    DEFAULT_LOCAL_DEVICE_NAME,
    DEFAULT_REMOTE_DEVICE_PORT,
)

"""
    TODO:
    - Create ~/.beam folder (contains API key config, syncthing binary, etc)
    - Download syncthing binary into ~/.beam folder, depending on OS/distro
"""

namespace = "luke-lombardi"
API_KEY = "testmykey"  # TODO: set api key to sandbox session ID dynamically


class SyncthingClient:
    def __init__(self, sandbox_session_id: str) -> None:
        self.sandbox_session_id: str = sandbox_session_id
        self.process_manager: ProcessManager = ProcessManager()
        self.local_client: Union[Syncthing, None] = None
        self.local_config: Union[dict, None] = None
        self.remote_client: Union[Syncthing, None] = None
        self.remote_config: Union[dict, None] = None
        self.max_connection_attempts: int = 10

        self.sync_api_host = (
            f"sandbox-{namespace}.slai.okteto.dev/sync-gui/{self.sandbox_session_id}/"
        )
        self.sync_host = (
            f"sandbox-{namespace}.slai.okteto.dev:443/sync/{self.sandbox_session_id}/"
        )

        if os.path.exists(f"{DEFAULT_CONFIG_DIRECTORY}/config.xml"):
            os.remove(f"{DEFAULT_CONFIG_DIRECTORY}/config.xml")

        self._start_local_server(port=DEFAULT_SYNCTHING_LOCAL_PORT)

        local_device_id = None
        remote_device_id = None

        # Connect to local syncthing server
        self.local_client, self.local_config = self._connect_to_server(
            host="0.0.0.0", port=DEFAULT_SYNCTHING_LOCAL_PORT
        )
        for device in self.local_config["devices"]:
            if device["name"] != DEFAULT_REMOTE_DEVICE_NAME:
                local_device_id = device["deviceID"]
                break

        # Connect to remote syncthing server (on runner)
        self.remote_client, self.remote_config = self._connect_to_server(
            host=self.sync_api_host, port=None, https=True
        )

        for device in self.remote_config["devices"]:
            if device["name"] != DEFAULT_LOCAL_DEVICE_NAME:
                remote_device_id = device["deviceID"]
                break

        # Add the workspace folder to share with the runner
        self.local_config["folders"] = generate_folder_template(path=os.getcwd())
        self.local_config["folders"][0]["devices"].append(
            generate_folder_device_template(device_id=local_device_id)
        )
        self.local_config["folders"][0]["devices"].append(
            generate_folder_device_template(device_id=remote_device_id)
        )

        if len(self.local_config["devices"]) == 1:
            self.local_config["devices"].append(
                generate_remote_device_template(
                    device_name="remote",
                    device_id=remote_device_id,
                    host=self.sync_host,
                    port=443,
                )
            )

        self.local_client.system.set_config(self.local_config, and_restart=False)

        # Update remote config
        if len(self.remote_config["devices"]) == 1:
            self.remote_config["devices"].append(
                generate_local_device_template(
                    device_name="local", device_id=local_device_id
                )
            )

        # Here, we override the default listening port (which is dynamic)
        self.remote_config["options"]["reconnectionIntervalS"] = 1

        # TODO: pretty sure we can just use the default port (22000)
        # which should improve startup speed slightly
        # self.remote_config["options"]["listenAddresses"] = [
        #     # f"tcp://0.0.0.0:{DEFAULT_REMOTE_DEVICE_PORT}",
        #     "dynamic+https://relays.syncthing.net/endpoint",
        #     # f"quic://0.0.0.0:{DEFAULT_REMOTE_DEVICE_PORT}",
        # ]

        self.remote_client.system.set_config(self.remote_config, and_restart=False)

    def _start_local_server(self, port: int = DEFAULT_SYNCTHING_LOCAL_PORT):
        """Start local syncthing server"""

        self.process_manager.start(
            [
                f"{DEFAULT_CONFIG_DIRECTORY}/syncthing",
                f"--home={DEFAULT_CONFIG_DIRECTORY}/",
                f"--gui-address=0.0.0.0:{port}",
                "--no-browser",
                f"--gui-apikey={API_KEY}",
                f"--logfile={DEFAULT_CONFIG_DIRECTORY}/syncthing.log",
                "--log-max-old-files=0",
            ]
        )

    def _connect_to_server(
        self, *, host: str, port: Union[int, None], https: bool = False
    ) -> Tuple[Syncthing, dict]:
        """Connect to syncthing server and pull initial configuration"""

        client = Syncthing(API_KEY, host=host, port=port, is_https=https)

        initial_config = None
        for _ in range(self.max_connection_attempts):
            try:
                initial_config = client.system.config()
                break
            except SyncthingError:
                time.sleep(0.5)
                continue

        if not initial_config:
            raise RuntimeError("unable_to_connect_to_syncthing")

        return (client, initial_config)



DEFAULT_FOLDER_NAME = "beam-1"
DEFAULT_REMOTE_DEVICE_NAME = "remote"
DEFAULT_LOCAL_DEVICE_NAME = "local"
DEFAULT_REMOTE_DEVICE_PORT = 22000


def generate_local_device_template(*, device_name: str, device_id: str) -> dict:
    return {
        "deviceID": device_id,
        "name": device_name,
        "addresses": ["dynamic"],
        "compression": "metadata",
        "certName": "",
        "introducer": True,
        "skipIntroductionRemovals": False,
        "introducedBy": "",
        "paused": False,
        "allowedNetworks": [],
        "autoAcceptFolders": True,
        "maxSendKbps": 0,
        "maxRecvKbps": 0,
        "ignoredFolders": [],
        "maxRequestKiB": 0,
        "untrusted": False,
        "remoteGUIPort": 0,
    }


def generate_remote_device_template(
    *,
    device_name: str,
    device_id: str,
    host: str = "0.0.0.0",
    port: int = DEFAULT_REMOTE_DEVICE_PORT,
) -> dict:
    return {
        "deviceID": device_id,
        "name": device_name,
        "addresses": [f"tcp://{host}"],
        "compression": "metadata",
        "certName": "",
        "introducer": False,
        "skipIntroductionRemovals": False,
        "introducedBy": "",
        "paused": False,
        "allowedNetworks": [],
        "autoAcceptFolders": False,
        "maxSendKbps": 0,
        "maxRecvKbps": 0,
        "ignoredFolders": [],
        "maxRequestKiB": 0,
        "untrusted": False,
        "remoteGUIPort": 0,
    }


def generate_folder_device_template(device_id: str) -> dict:
    return {
        "deviceID": device_id,
        "introducedBy": "",
        "encryptionPassword": "",
    }


def generate_folder_template(path: str):
    return [
        {
            "id": DEFAULT_FOLDER_NAME,
            "label": DEFAULT_FOLDER_NAME,
            "filesystemType": "basic",
            "path": path,
            "type": "sendreceive",
            "devices": [],
            "rescanIntervalS": 300,
            "fsWatcherEnabled": True,
            "fsWatcherDelayS": 1,
            "ignorePerms": False,
            "autoNormalize": True,
            "minDiskFree": {"value": 1, "unit": "%"},
            "versioning": {
                "type": "",
                "params": {},
                "cleanupIntervalS": 3600,
                "fsPath": "",
                "fsType": "basic",
            },
            "copiers": 0,
            "pullerMaxPendingKiB": 0,
            "hashers": 0,
            "order": "random",
            "ignoreDelete": False,
            "scanProgressIntervalS": 1,
            "pullerPauseS": 0,
            "maxConflicts": 0,
            "disableSparseFiles": False,
            "disableTempIndexes": False,
            "paused": False,
            "weakHashThresholdPct": 25,
            "markerName": ".",
            "copyOwnershipFromParent": False,
            "modTimeWindowS": 0,
            "maxConcurrentWrites": 2,
            "disableFsync": False,
            "blockPullOrder": "standard",
            "copyRangeMethod": "all",
            "caseSensitiveFS": False,
            "junctionsAsDirs": False,
        }
    ]
